import pytest

from tests.testsupport import *


def test_bytearray_malloc():
    assert len(sabctools.bytearray_malloc(10)) == 10


def test_bytearray_malloc_bad_inputs():
    with pytest.raises(TypeError):
        sabctools.bytearray_malloc(10.0)
    with pytest.raises(SystemError):
        sabctools.bytearray_malloc(-1)
    with pytest.raises(TypeError):
        sabctools.bytearray_malloc("foo")



import time
import ssl
import sabctools
import socket
import time

hostname = "eunews.frugalusenet.com"
context = ssl.create_default_context()

print(sabctools.openssl_linked)

buffer = bytearray(100)
bufferview = memoryview(buffer)

with socket.create_connection((hostname, 563)) as sock:
    with context.wrap_socket(sock, server_hostname=hostname) as ssock:
        ssock.setblocking(False)
        time.sleep(1)
        print(ssock.version())
        print(sabctools.unlocked_ssl_recv_into(ssock._sslobj, bufferview[99:]))
        print(buffer)



import io
import os
import sys

import pytest
import glob
from tests.testsupport import *


@pytest.mark.parametrize(
    "filename",
    ["test_regular.yenc", "test_regular_2.yenc"],
)
def test_regular(filename: str):
    data_plain = read_plain_yenc_file(filename)
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


def test_partial():
    data_plain = read_plain_yenc_file("test_partial.yenc")
    decoded_data, filename, filesize, begin, size, crc_correct = sabctools_yenc_wrapper(data_plain)
    assert filename == "90E2Sdvsmds0801dvsmds90E.part06.rar"
    assert filesize == 49152000
    assert begin == 15360000
    assert size == 384000
    assert crc_correct is None
    assert len(decoded_data) == 549


def test_special_chars():
    data_plain = read_plain_yenc_file("test_special_chars.yenc")
    # We only compare the data and the filename
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)

    data_plain = read_plain_yenc_file("test_special_utf8_chars.yenc")
    # We only compare the data and the filename
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


def test_bad_crc():
    data_plain = read_plain_yenc_file("test_bad_crc.yenc")
    # We only compare the data and the filename
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


def test_bad_crc_end():
    data_plain = read_plain_yenc_file("test_bad_crc_end.yenc")
    _, _, _, _, _, crc = sabctools_yenc_wrapper(data_plain)
    assert crc is None


def test_no_filename():
    data_plain = read_plain_yenc_file("test_no_name.yenc")
    _, filename, _, _, _, _ = sabctools_yenc_wrapper(data_plain)
    assert filename is None


def test_padded_crc():
    data_plain = read_plain_yenc_file("test_padded_crc.yenc")
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


@pytest.mark.parametrize(
    "filename",
    [
        "test_end_after_filename.yenc",
        "test_end_after_ypart.yenc",
    ],
)
def test_end_after(filename: str):
    data_plain = read_plain_yenc_file(filename)
    decoded_data, _, _, _, _, _ = sabctools_yenc_wrapper(data_plain)
    assert decoded_data is None


def test_ref_counts():
    """Note that sys.getrefcount itself adds another reference!"""
    # In Python 3.14+, getrefcount returns 1, in earlier versions it returns 2
    expected_refcount = 1 if sys.version_info >= (3, 14) else 2

    # Test regular case
    data_plain = read_plain_yenc_file("test_regular.yenc")
    data_out, filename, filesize, begin, end, crc_correct = sabctools_yenc_wrapper(data_plain)

    assert sys.getrefcount(data_plain) == expected_refcount
    assert sys.getrefcount(data_out) == expected_refcount
    assert sys.getrefcount(filename) == expected_refcount
    assert sys.getrefcount(begin) == expected_refcount
    assert sys.getrefcount(end) == expected_refcount
    assert sys.getrefcount(crc_correct) == expected_refcount

    # Test further processing
    data_plain = read_plain_yenc_file("test_bad_crc_end.yenc")
    _, _, _, _, _, crc = sabctools_yenc_wrapper(data_plain)
    assert crc is None
    assert sys.getrefcount(data_plain) == expected_refcount


@pytest.mark.parametrize(
    "filename",
    sorted(glob.glob("tests/yencfiles/crc_*")),
)
def test_crc_pickles(filename: str):
    data_plain = read_pickle(filename)
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


@pytest.mark.parametrize(
    "filename",
    sorted(glob.glob("tests/yencfiles/small_file*")),
)
def test_small_file_pickles(filename: str):
    data_plain = read_pickle(filename)
    assert python_yenc(data_plain) == sabctools_yenc_wrapper(data_plain)


@pytest.mark.parametrize(
    "code",
    [
        # Article
        223,  # stat
        412,  # No newsgroup selected
        423,  # No article with that number
        420,  # No newsgroup selected
        430,  # Article not found
        # Auth
        281,  # Authentication accepted
        381,  # Password required
        481,  # Authentication failed/rejected
        482,  # Authentication commands issued out of sequence
        # Generic
        500,  # Unknown command
        501,  # Syntax error
        502,  # Command unavailable
        503,  # Not supported
    ],
)
def test_nntp_not_multiline(code: int):
    line = bytes(f"{code} 0 <message-id>\r\n", encoding="utf-8")
    input = BytesIO(line)
    decoder = sabctools.Decoder(len(line))
    n = input.readinto(decoder)
    decoder.process(n)
    response = next(decoder, None)
    assert response
    assert response.data is None
    assert response.status_code == code


def test_head():
    data_plain = read_plain_yenc_file("test_head.yenc")
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)

    response = next(decoder, None)
    assert response
    assert response.data is None
    assert response.status_code == 221
    assert response.lines is not None
    assert len(response.lines) == 13
    assert "X-Received-Bytes: 740059" in response.lines


def test_capabilities():
    data_plain = read_plain_yenc_file("capabilities.yenc")
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)

    response = next(decoder, None)
    assert response
    assert response.data is None
    assert len(response.lines) == 2
    assert "VERSION 1" in response.lines
    assert "AUTHINFO USER PASS" in response.lines


def test_article():
    data_plain = read_plain_yenc_file("test_article.yenc")
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)

    response = next(decoder, None)
    assert response
    assert response.data
    assert response.status_code == 220
    assert response.lines is not None
    assert len(response.lines) == 13
    assert "X-Received-Bytes: 740059" in response.lines
    assert len(response.data) == 716800


def test_streaming():
    BUFFER_SIZE = 1024
    yenc_files = ["test_regular_2.yenc"] * 5 + ["test_special_utf8_chars.yenc"]
    responses = []

    # Read in chunks like a network
    input = io.BytesIO()
    for filename in yenc_files:
        input.write(read_plain_yenc_file(filename))
    input.seek(0)

    decoder = sabctools.Decoder(BUFFER_SIZE)
    while (n := input.readinto(decoder)) != 0:
        decoder.process(n)

    for response in decoder:
        responses.append(response)

    assert len(responses) == len(yenc_files)

    for i, dec in enumerate(responses):
        assert dec.status_code in (220, 222)
        assert python_yenc(read_plain_yenc_file(yenc_files[i])) == (
            dec.data,
            correct_unknown_encoding(dec.file_name),
            dec.file_size,
            dec.part_begin,
            dec.part_size,
            dec.crc,
        )


def test_uu():
    data_plain = read_uu_file("logo_full.nntp")
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)
    response = next(decoder, None)
    assert response
    assert response.data
    assert response.lines is None
    assert response.file_name == "logo-full.svg"
    assert response.file_size == 2184
    assert response.crc == 0x6BC2917D


@pytest.mark.parametrize(
    "length",
    range(1, 46),
)
def test_uu_length(length: int):
    expected = os.urandom(length)
    parts = [b"222 0 <foo@bar>\r\n", uu(expected), b"\r\n" b".\r\n"]
    data_plain = b"".join(parts)
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)
    response = next(decoder, None)
    assert response
    assert response.format is sabctools.EncodingFormat.UU
    assert response.bytes_decoded
    assert response.data == expected


# Tests for super-invalid inputs to ensure decoder doesn't crash


@pytest.mark.parametrize(
    "filename",
    [
        # Protocol/Status edge cases
        "test_invalid_status_code.yenc",  # Non-numeric status code
        "test_truncated_status.yenc",  # Incomplete status line
        "test_empty_file.yenc",  # Empty file
        "test_only_newlines.yenc",  # Only newlines
        # Malformed yEnc headers
        "test_malformed_ybegin.yenc",  # ybegin missing required fields
        "test_negative_size.yenc",  # Negative size value
        "test_huge_size.yenc",  # Extremely large size
        "test_huge_size_1TiB.yenc",  # Extremely large size (1 TB)
        "test_double_ybegin.yenc",  # Two ybegin lines
        # Structure violations
        "test_missing_yend.yenc",  # ybegin without yend
        "test_ypart_without_ybegin.yenc",  # ypart before ybegin
        "test_ypart_invalid_range.yenc",  # ypart begin > end
        "test_part_exceeds_limit.yenc",  # Part size > 10MB limit
        # Special characters & encoding
        "test_non_ascii_everywhere.yenc",  # UTF-8/Chinese characters
        "test_only_dots.yenc",  # Dot-stuffing edge case
        "test_invalid_escape.yenc",  # Invalid escape sequences
        # CRC edge cases (tested separately due to additional assertions)
        "test_extremely_long_crc.yenc",  # CRC exceeding 64-bit
    ],
)
def test_invalid_inputs_no_crash(filename: str):
    """Test that decoder handles super-invalid inputs gracefully without crashing."""
    data_plain = read_plain_yenc_file(filename)
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)

    # Basic check: decoder should not crash
    assert decoder is not None


def test_exceeds_size_limit():
    size = 20 * 1024 * 1024
    output, crc = sabctools.yenc_encode(b"\x00" * size)
    parts = [
        b"222 0 <foo@bar>\r\n" b"=ybegin part=1 total=1 line=128 size=%d name=helloworld\r\n" % size,
        b"=ypart begin=1 end=%d\r\n" % size,
        output,
        b"\r\n=yend size=%d pcrc32=%s\r\n" % (size, hex(crc).encode()[2:]),
        b".\r\n",
    ]
    data_plain = b"".join(parts)
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(256 * 1024)
    with pytest.raises(BufferError, match="Maximum data buffer size exceeded"):
        while n := input.readinto(decoder):
            decoder.process(n)


def test_invalid_crc_chars():
    """Test with non-hex characters in CRC field - crc_expected should be None."""
    data_plain = read_plain_yenc_file("test_invalid_crc_chars.yenc")
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)
    response = next(decoder, None)
    assert response
    assert response.crc_expected is None


@pytest.mark.parametrize(
    "hex,expected",
    [
        ["ffffffffa95d3e50", 0xA95D3E50],
        ["fffffffa95d3e50", 0xA95D3E50],
        ["ffffffa95d3e50", 0xA95D3E50],
        ["fffffa95d3e50", 0xA95D3E50],
        ["ffffa95d3e50", 0xA95D3E50],
        ["fffa95d3e50", 0xA95D3E50],
        ["ffa95d3e50", 0xA95D3E50],
        ["fa95d3e50", 0xA95D3E50],
        ["a95d3e50", 0xA95D3E50],
        ["a95d3e5", 0xA95D3E5],
        ["a95d3e", 0xA95D3E],
        ["a95d3", 0xA95D3],
        ["a95d", 0xA95D],
        ["a95", 0xA95],
        ["a9", 0xA9],
        ["a", 0xA],
        ["", 0],
        ["12345678 ", 0x12345678],  # space at end
    ],
)
def test_parsing_crc(hex: str, expected: int):
    parts = [
        b"222 0 <foo@bar>\r\n"
        b"=ybegin part=1 total=1 line=128 size=12 name=helloworld\r\n"
        b"=ypart begin=1 end=12\r\n"
        b"r\x8f\x96\x96\x99J\xa1\x99\x9c\x96\x8eK\r\n"
        b"=yend size=12 pcrc32=%s\r\n" % hex.encode(),
        b".\r\n",
    ]
    data_plain = b"".join(parts)
    input = BytesIO(data_plain)
    decoder = sabctools.Decoder(len(data_plain))
    n = input.readinto(decoder)
    decoder.process(n)
    response = next(decoder, None)
    assert response
    assert response.crc_expected == expected



#!/usr/bin/python3 -OO
# Copyright 2007-2019 The SABnzbd-Team (sabnzbd.org)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

###################
# SUPPORT FUNCTIONS
###################
from zlib import crc32
import re
import pickle
from typing import Tuple, Optional, Union
from io import BytesIO
import binascii

import chardet
import sabctools


def correct_unknown_encoding(str_or_bytes_in: Union[str, bytes]) -> str:
    """Files created on Windows but unpacked/repaired on
    linux can result in invalid filenames. Try to fix this
    encoding by going to bytes and then back to unicode again.
    Last resort we use chardet package
    """
    # If already string, back to bytes
    if not isinstance(str_or_bytes_in, bytes):
        str_or_bytes_in = str_or_bytes_in.encode("utf-8", "surrogateescape")

    # Try simple bytes-to-string
    try:
        return str_or_bytes_in.decode("utf-8")
    except UnicodeDecodeError:
        try:
            # Try using 8-bit ASCII, if came from Windows
            return str_or_bytes_in.decode("ISO-8859-1")
        except ValueError:
            # Last resort we use the slow chardet package
            return str_or_bytes_in.decode(chardet.detect(str_or_bytes_in)["encoding"])


def read_plain_yenc_file(filename: str) -> bytearray:
    with open("tests/yencfiles/%s" % filename, "rb") as yencfile:
        return bytearray(yencfile.read())


def read_uu_file(filename: str) -> bytearray:
    with open("tests/uufiles/%s" % filename, "rb") as uufile:
        return bytearray(uufile.read())


def read_pickle(filename):
    with open(filename, "rb") as yencfile:
        try:
            data_chunks, data_bytes = pickle.load(yencfile, encoding="bytes")
        except:
            # Reset the pointer and try again
            yencfile.seek(0)
            data_chunks, data_bytes, lines = pickle.load(yencfile, encoding="bytes")
    return bytearray(b"".join(data_chunks))


def sabctools_yenc_wrapper(data: bytearray) -> Tuple[bytearray, str, int, int, int, Optional[int]]:
    input = BytesIO(data)
    decoder = sabctools.Decoder(len(data))
    assert not decoder
    assert not len(decoder)
    n = input.readinto(decoder)
    decoder.process(n)
    assert decoder
    assert len(decoder)
    response = next(decoder)
    assert response
    assert response.status_code in (220, 222)
    assert response.lines is None
    assert response.format is sabctools.EncodingFormat.YENC
    assert response.message

    return response.data, response.file_name, response.file_size, response.part_begin, response.part_size, response.crc


def python_yenc(data_plain):
    """Use the older decoder to verify the new one"""
    data = []

    # Remove the NNTP-double-dot style
    new_lines = data_plain.split(b"\r\n")
    for i in range(len(new_lines)):
        if new_lines[i][:2] == b"..":
            new_lines[i] = new_lines[i][1:]
    if new_lines[-1] == b".":
        new_lines = new_lines[1:-1]
    data.extend(new_lines)

    # Parse the yEnc headers
    yenc, data = parse_yenc_data(data)
    ybegin, ypart, yend = yenc

    # Now we get the true flat data
    flat_yenc_data = b"".join(data)

    # Remove the escaped-chars
    for i in (0, 9, 10, 13, 27, 32, 46, 61):
        j = b"=%c" % (i + 64)
        flat_yenc_data = flat_yenc_data.replace(j, b"%c" % i)

    # Use the much faster translate function to do fast-subtract of 42
    from_bytes = b"".join([b"%c" % i for i in range(256)])
    to_bytes = b"".join([b"%c" % ((i + 256 - 42) % 256) for i in range(256)])
    translate_table = bytes.maketrans(from_bytes, to_bytes)
    decoded_data = flat_yenc_data.translate(translate_table)

    # Detect begin and ending
    begin = end = size = 0
    if ybegin:
        # not multi-part
        if not ybegin.get("part"):
            size = int(ybegin.get("size"))
    if ypart:
        if begin := ypart.get("begin"):
            begin = int(begin)
        if end := ypart.get("end"):
            end = int(end)
        if end and begin:
            size = end - begin + 1
            begin -= 1

    return decoded_data, ybegin["name"], int(ybegin["size"]), begin, size, crc32(decoded_data)


def parse_yenc_data(data):
    ybegin = None
    ypart = None
    yend = None

    # Check head
    for i in range(min(40, len(data))):
        try:
            if data[i].startswith(b"=ybegin "):
                splits = 3
                if data[i].find(b" part=") > 0:
                    splits += 1
                if data[i].find(b" total=") > 0:
                    splits += 1

                ybegin = get_yenc_data(data[i], splits)

                if data[i + 1].startswith(b"=ypart "):
                    ypart = get_yenc_data(data[i + 1])
                    data = data[i + 2 :]
                    break
                else:
                    data = data[i + 1 :]
                    break
        except IndexError:
            break

    # Check tail
    for i in range(-1, -11, -1):
        try:
            if data[i].startswith(b"=yend "):
                yend = get_yenc_data(data[i])
                data = data[:i]
                break
        except IndexError:
            break

    return ((ybegin, ypart, yend), data)


def get_yenc_data(line, splits=None):
    # Example: =ybegin part=1 line=128 size=123 name=-=DUMMY=- abc.par
    YSPLIT_RE = re.compile(b"([a-zA-Z0-9]+)=")

    fields = {}

    if splits:
        parts = YSPLIT_RE.split(line, splits)[1:]
    else:
        parts = YSPLIT_RE.split(line)[1:]

    if len(parts) % 2:
        return fields

    for i in range(0, len(parts), 2):
        key, value = parts[i], parts[i + 1]
        fields[correct_unknown_encoding(key)] = correct_unknown_encoding(value.strip())

    return fields


def yenc_subtract(char, subtract):
    """Wrap-around for below 0"""
    char_diff = char - subtract
    if char_diff < 0:
        return 256 + char_diff
    return char_diff


def uu(data: bytes):
    """Uuencode data and insert a period if necessary"""
    line = binascii.b2a_uu(data).rstrip(b"\n")

    # Dot stuffing
    if line.startswith(b"."):
        return b"." + line

    return line



from tests.testsupport import *


def test_encoder():
    output, crc = sabctools.yenc_encode(b"Hello world!")
    assert output == b"r\x8f\x96\x96\x99J\xa1\x99\x9c\x96\x8eK"
    assert crc == 0x1B851995



import os
import select
import socket
import ssl
import sys
import tempfile
import threading
import time

import pytest
import portend

from tests.testsupport import *

HOST = "127.0.0.1"

cert = """-----BEGIN CERTIFICATE-----
MIIFazCCA1OgAwIBAgIUXYYp1atAnSypLyKprYna65kFn9owDQYJKoZIhvcNAQEL
BQAwRTELMAkGA1UEBhMCQVUxEzARBgNVBAgMClNvbWUtU3RhdGUxITAfBgNVBAoM
GEludGVybmV0IFdpZGdpdHMgUHR5IEx0ZDAeFw0yMzAxMTAwODE5NDVaFw0zMzAx
MDcwODE5NDVaMEUxCzAJBgNVBAYTAkFVMRMwEQYDVQQIDApTb21lLVN0YXRlMSEw
HwYDVQQKDBhJbnRlcm5ldCBXaWRnaXRzIFB0eSBMdGQwggIiMA0GCSqGSIb3DQEB
AQUAA4ICDwAwggIKAoICAQDWHsTQOLJ0O/WdJ2pr/IjwXEcrFr6xiDuxQGwJQxkI
beY4w/jFxz9TmRwkIm+wrqw6o/4BxRJIJrot0byaJzSYlYJS+ChH1rRa3E72Jb0L
Kfk1peNcTInujHStXv93aLbxhamswg2M7gTUPelZeIlTHf/YsLzx4KEfF/PSs6DJ
eOUHmGcFE01Ypef0tM+72DncrRbDWKbAC5IQm2g4yCC/L6RYNbpOn5Cm9/M1JPnx
+Gn6ylrWQhquVrLnxEfupNgCWiRTZXSvy6jNPgGEi92OiwT5M3MMn/FZVxDKILHg
B9esr1b3s6qTW3sclcLXKIC1eG1/ikUmVfbVonmnlLuHao6SfHWeV7kuBdqnYQFN
UgZ11xtakWC/dwjIt101MjFe3PlgLfgxeSMXZFYpBc0RwBfmBavEHbbRTgebloIO
xrPMIvbFiNg9osoC8nTKwtrEMo0MlCYyElXs18hLzo3GvL1bCUCQiqj4sRBFKG4B
SgyhV/i72kuf9FhdcM4IJFoCc9B2Ll9KCflhhwGe+f/1cpd6nlKXgI2+UfgHY0Nj
UqDqq7gLg0WMCpSKasH68e4UDzceUUB1d7OwJwljxJxGuZErKIQbGi0wKVy0evRO
lcb5IL9l8rPfxE1aqlQ6VbDJ91rOrYdB6MugWNK3KfYPcAXHkPsWcMPtO4R2nEeR
FwIDAQABo1MwUTAdBgNVHQ4EFgQUp9LcqyRllH3ZN/WpSPjxVybPSZowHwYDVR0j
BBgwFoAUp9LcqyRllH3ZN/WpSPjxVybPSZowDwYDVR0TAQH/BAUwAwEB/zANBgkq
hkiG9w0BAQsFAAOCAgEAaeyercYN95kMCdLCjX9jmAGhG6WB3eFvEwmTtcSoxo3Q
6/LBk8EdikfirIHuJKZVxTRuvRA/KniDI/pJC16qcRUI0OmQFtCirqD7kiR/fkYO
IXqh3wzu4bBGewf+MLGtJjYxn83w/KP7kj3ZTaoOLbmizBDWCMlYdYwbqTVfuOCL
hQwvRpiqeOBNCphvrreo6HqLaiWaRPb60NZqEu+yI7ltGN+QDKQJ04CJT1gLqkAB
MsmvLLmQDPtnhE16zqZc5iZ3FHFm+sSSQq5jImjmCHT8aNGk4ng0VuQQsMZClgmx
U6SvMo2u5akK9h3d9L2QT9vc8Zn/gUdQW4N1kLBU2HoHod/KLKkPrWGR4oIowYmk
C/nQjh/qxJFk6zEwo7Woo2+mQop9vKnA5jH24qbuCm0k4GfJegB7FCaW7TcRTwxy
wExvMNVgaZ3+Fuf/dbGTG3EHSfQO+hPMFCvo7WQ+ndQcqEmLlyNnV9owfHOYYo1l
TgLUi49uVqHLfN3BZPIOVW3FP26OLxdVRQf1OdwR5+eDp6CXceO5U7kZJN5L7ZMi
fDIh4KerzRZrtdt6Y/CnxI6R1qDHfpK2BA8DJpyhhDUap7OrGhktqb1nTT0hrLAq
acN1EC6NLepwbQL0RR3Xhmu3Jq59bXRVpjCMHC4rQHQvbG01vT0SjENM38RGUbk=
-----END CERTIFICATE-----
"""

key = """-----BEGIN PRIVATE KEY-----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-----END PRIVATE KEY-----
"""


class EchoServer(threading.Thread):
    """An SSL echo server"""

    class ConnectionHandler(threading.Thread):
        def __init__(self, server, connsock, addr, buffer_size):
            self.server = server
            self.sock = connsock
            self.addr = addr
            self.buffer_size = buffer_size
            self.running = False
            self.sslconn = None
            super().__init__()

        def wrap_conn(self):
            try:
                self.sslconn = self.server.context.wrap_socket(self.sock, server_side=True)
            except (ConnectionResetError, BrokenPipeError, ConnectionAbortedError) as e:
                self.running = False
                self.close()
                return False
            except (ssl.SSLError, OSError):
                return False
            return True

        def read(self):
            return self.sslconn.read(self.buffer_size)

        def write(self, bytes: bytes):
            self.sslconn.write(bytes)

        def close(self):
            if self.sslconn:
                self.sslconn.close()

        def run(self) -> None:
            self.running = True
            if not self.wrap_conn():
                return
            while self.running:
                try:
                    data = self.read()
                    if not data.strip():
                        self.running = False
                        try:
                            self.sock = self.sslconn.unwrap()
                        except OSError:
                            # Incorrect shutdown of SSL session
                            pass
                        else:
                            self.sslconn = None
                        self.close()
                    else:
                        self.write(data)
                except OSError:
                    self.close()
                    self.running = False
                    self.server.stop()

    def __init__(self):
        self.host = HOST
        self.port = portend.find_available_local_port()
        self.flag = None
        self.active = False
        super().__init__()

        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            self.certfile = tmp.name
            with open(self.certfile, "w") as f:
                f.write(cert)
                f.close()

        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            self.keyfile = tmp.name
            with open(self.keyfile, "w") as f:
                f.write(key)
                f.close()

        self.context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.host, self.port))

    def __enter__(self):
        self.start(threading.Event())
        self.flag.wait()
        return self

    def __exit__(self, *args):
        self.stop()
        self.join()

    def start(self, flag=None):
        self.flag = flag
        threading.Thread.start(self)

    def run(self) -> None:
        self.sock.settimeout(2)
        self.sock.listen(5)
        self.active = True
        if self.flag:
            self.flag.set()
        while self.active:
            try:
                conn, addr = self.sock.accept()
                handler = self.ConnectionHandler(self, conn, addr, 131072)
                handler.start()
                handler.join()
            except TimeoutError:
                pass
            except KeyboardInterrupt:
                self.stop()

        self.close()

    def close(self):
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        os.unlink(self.certfile)
        os.unlink(self.keyfile)

    def stop(self):
        self.active = False


@pytest.fixture()
def server():
    server = EchoServer()
    with server:
        yield server


@pytest.fixture()
def client(server):
    with socket.create_connection((server.host, server.port)) as sock:
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        context.check_hostname = False
        context.verify_mode = ssl.VerifyMode.CERT_NONE
        with context.wrap_socket(sock, server_hostname=server.host) as ssock:
            ssock.setblocking(False)
            yield ssock


def unlocked_ssl_recv_into_wrapper(sock: ssl.SSLSocket, data: bytes, data_view: memoryview) -> int:
    """Send data to the socket and writes the response to a buffer"""
    sock.sendall(data)
    data_position = 0
    while True:
        try:
            data_position += sabctools.unlocked_ssl_recv_into(sock, data_view[data_position:])
        except ssl.SSLWantReadError:
            pass
        if data_position > 0:
            break
    return data_position


@pytest.fixture()
def buffer():
    yield memoryview(bytearray(1024))


def test_openssl_linked():
    assert sabctools.openssl_linked is True


def test_unlocked_ssl_recv_into_not_a_socket_fails(buffer):
    with pytest.raises(TypeError, match=r"argument 1 must be SSLSocket"):
        sabctools.unlocked_ssl_recv_into("this is not a socket", buffer)


def test_unlocked_ssl_recv_into_not_a_buffer_fails(client):
    with pytest.raises(TypeError, match=r"argument 2 must be read-write bytes-like object"):
        unlocked_ssl_recv_into_wrapper(client, b"TEST", "...")


def test_unlocked_ssl_recv_into_response(client, buffer):
    bytes_received = unlocked_ssl_recv_into_wrapper(client, b"TEST", buffer)
    assert buffer[:bytes_received].tobytes() == b"TEST"


def test_unlocked_ssl_recv_into_bulk_response(client):
    # 131072 bytes divide up into 8 TLS records (16 KB each)
    # In nonblocking mode, we should be able to read all eight in a single
    # drop of the GIL.
    size = 131072
    buffer = bytearray(size)

    client.sendall(b"\xff" * size)

    select.select([client], [], [])

    while size > 0:
        try:
            count = sabctools.unlocked_ssl_recv_into(client, buffer)
        except ssl.SSLWantReadError:
            select.select([client], [], [])
            # Give the sender some more time to complete sending.
            time.sleep(0.1)
        else:
            if count > 16384:
                return
            size -= count

    pytest.fail("All TLS reads were smaller than 16KB")


#
def test_unlocked_ssl_recv_into_blocking_socket_fails(client, buffer):
    with pytest.raises(ValueError, match="Only non-blocking sockets are supported"):
        client.setblocking(True)
        unlocked_ssl_recv_into_wrapper(client, b"TEST", buffer)


def test_unlocked_ssl_recv_into_full_buffer_fails(client, buffer):
    with pytest.raises(ValueError, match="No space left in buffer"):
        unlocked_ssl_recv_into_wrapper(client, b"TEST", buffer[len(buffer) :])


def test_unlocked_ssl_recv_into_ref_counts_unchanged(client, buffer):
    objects = [
        client,
        client._sslobj,
        buffer,
    ]

    before = [sys.getrefcount(x) for x in objects]

    bytes_received = unlocked_ssl_recv_into_wrapper(client, b"Hello World", buffer)
    assert buffer[:bytes_received].tobytes() == b"Hello World"

    after = [sys.getrefcount(x) for x in objects]

    assert after == before



import os
import subprocess
import sys
import tempfile
import pytest
from typing import IO

from tests.testsupport import *


def test_sparse():
    file = tempfile.NamedTemporaryFile(delete=False)
    try:
        sabctools.sparse(file, 100)
        assert os.path.getsize(file.name) == 100
        assert is_sparse(file) is True
    finally:
        file.close()
        os.unlink(file.name)


def test_sparse_fd():
    file = tempfile.NamedTemporaryFile(delete=False)
    try:
        sabctools.sparse(file.fileno(), 100)
        assert os.path.getsize(file.name) == 100
        assert is_sparse(file) is True
    finally:
        file.close()
        os.unlink(file.name)


@pytest.mark.parametrize(
    "length,position",
    [
        (1024, 0),
        (1024, 512),
        (1024, 4096),
    ],
)
def test_sparse_position_expected(length: int, position: int):
    with tempfile.TemporaryFile() as file:
        file.write(b"Hello World!")
        file.seek(position)
        sabctools.sparse(file, length)
        assert os.stat(file.name).st_size == length
        assert file.tell() == position


def is_sparse(file: IO) -> bool:
    """Check if a path is a sparse file"""
    stat = os.stat(file.name)
    if "win32" in sys.platform:
        return bool(stat.st_file_attributes & 0x200)

    # Linux and macOS
    if stat.st_blocks * 512 < stat.st_size:
        return True

    # Filesystem with SEEK_HOLE (ZFS)
    try:
        with open(file.name, "rb") as f:
            pos = f.seek(0, os.SEEK_HOLE)
            return pos < stat.st_size
    except (AttributeError, OSError):
        pass

    return False



#!/usr/bin/python3 -OO
# Copyright 2007-2023 The SABnzbd-Team (sabnzbd.org)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import platform
import re
import tempfile
import shutil
from setuptools import distutils
from distutils.ccompiler import CCompiler
from distutils.errors import CompileError
from typing import Type
from contextlib import closing, ExitStack

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils import log


def autoconf_check(
    compiler: Type[CCompiler], include_check: str = None, define_check: str = None, flag_check: str = None
):
    """A makeshift Python version of the autoconf checks"""
    with ExitStack() as stack:
        tmpdir = tempfile.mkdtemp()
        stack.callback(shutil.rmtree, tmpdir)
        (tmp_fd, tmp_path) = tempfile.mkstemp(suffix=".cc", prefix="sabctools_", dir=tmpdir)

        with closing(os.fdopen(tmp_fd, "w")) as f:
            if include_check:
                log.info("==> Checking support for include: %s", include_check)
                f.write(f"#include <{include_check}>\n")

            if define_check:
                log.info("==> Checking support for define: %s", define_check)
                # Just let it crash
                f.write(f"#ifndef {define_check}\n")
                f.write(f"#error {define_check} not available!\n")
                f.write(f"#endif\n")

            f.write("int main (int argc, char **argv) { return 0; }")

        extra_postargs = []
        if flag_check:
            log.info("==> Checking support for flag: %s", flag_check)
            extra_postargs.append(flag_check)

        try:
            log.info("==> Please ignore any errors shown below!")
            compiler.compile([tmp_path], output_dir=tmpdir, extra_postargs=extra_postargs)
            log.info("==> Success!")
        except CompileError:
            log.info("==> Not available!")
            return False

    return True


class SABCToolsBuild(build_ext):
    def build_extension(self, ext: Extension):
        # Try to determine the architecture to build for
        machine = platform.machine().lower()
        IS_X86 = machine in ["i386", "i686", "x86", "x86_64", "x64", "amd64"]
        IS_MACOS = sys.platform == "darwin"
        IS_ARM = machine.startswith("arm") or machine.startswith("aarch64")
        IS_AARCH64 = True

        log.info("==> Baseline detection: ARM=%s, x86=%s, macOS=%s", IS_ARM, IS_X86, IS_MACOS)

        # Determine compiler flags
        gcc_arm_neon_flags = []
        gcc_arm_crc_flags = []
        gcc_arm_crc_pmull_flags = []
        gcc_vpclmulqdq_flags = []
        gcc_vbmi2_flags = []
        gcc_avx10_flags = []
        gcc_rvv_flags = []
        gcc_rvzbkc_flags = []
        gcc_macros = []
        if self.compiler.compiler_type == "msvc":
            # LTCG not enabled due to issues seen with code generation where
            # different ISA extensions are selected for specific files
            ldflags = ["/OPT:REF", "/OPT:ICF"]
            cflags = ["/O2", "/GS-", "/Gy", "/sdl-", "/Oy", "/Oi"]
            if autoconf_check(self.compiler, flag_check="/std:c++20"):
                cflags.append("/std:c++20")
                ext.extra_compile_args.append("/std:c++20")
            elif autoconf_check(self.compiler, flag_check="/std:c++17"):
                cflags.append("/std:c++17")
                ext.extra_compile_args.append("/std:c++17")
            else:
                log.info("==> C++17 flag not available")
        else:
            # TODO: consider -flto - may require some extra testing
            ldflags = ["-ldl"]  # for dlopen
            cflags = [
                "-Wall",
                "-Wextra",
                "-Wno-unused-function",
                "-fomit-frame-pointer",
                "-fno-rtti",
                "-fno-exceptions",
                "-O3",
                "-fPIC",
                "-fwrapv",
            ]
            if autoconf_check(self.compiler, flag_check="-std=c++20"):
                cflags.append("-std=c++20")
                ext.extra_compile_args.append("-std=c++20")
            elif autoconf_check(self.compiler, flag_check="-std=c++17"):
                cflags.append("-std=c++17")
                ext.extra_compile_args.append("-std=c++17")
            else:
                log.info("==> C++17 flag not available")

            # Verify specific flags for ARM chips
            # macOS ARM does not need any flags, they support everything
            if IS_ARM and not IS_MACOS:
                if not autoconf_check(self.compiler, define_check="__aarch64__"):
                    log.info("==> __aarch64__ not available, disabling 64bit extensions")
                    IS_AARCH64 = False
                if autoconf_check(self.compiler, flag_check="-march=armv8-a+crc+crypto"):
                    gcc_arm_crc_pmull_flags.append("-march=armv8-a+crc+crypto")
                if autoconf_check(self.compiler, flag_check="-march=armv8-a+crc"):
                    gcc_arm_crc_flags.append("-march=armv8-a+crc")
                    # Resolve problems on armv7, see issue #56
                    if not IS_AARCH64:
                        gcc_arm_crc_flags.append("-fno-lto")
                        gcc_arm_crc_pmull_flags.append("-fno-lto")
                if not IS_AARCH64 and autoconf_check(self.compiler, flag_check="-mfpu=neon"):
                    gcc_arm_neon_flags.append("-mfpu=neon")
                    # Resolve problems on armv7, see issue #56
                    gcc_arm_neon_flags.append("-fno-lto")

            # Check for special x32 case
            if (
                IS_X86
                and not IS_MACOS
                and autoconf_check(self.compiler, define_check="__ILP32__")
                and autoconf_check(self.compiler, define_check="__x86_64__")
            ):
                log.info("==> Detected x32 platform, setting CRCUTIL_USE_ASM=0")
                ext.define_macros.append(("CRCUTIL_USE_ASM", "0"))
                gcc_macros.append(("CRCUTIL_USE_ASM", "0"))

            if IS_X86 and autoconf_check(self.compiler, flag_check="-mvpclmulqdq"):
                gcc_vpclmulqdq_flags = ["-mavx2", "-mvpclmulqdq", "-mpclmul"]

            if IS_X86 and autoconf_check(self.compiler, flag_check="-mavx512vbmi2"):
                gcc_vbmi2_flags = [
                    "-mavx512vbmi2",
                    "-mavx512vl",
                    "-mavx512bw",
                    "-mpopcnt",
                    "-mbmi",
                    "-mbmi2",
                    "-mlzcnt",
                ]

            if IS_X86 and autoconf_check(self.compiler, flag_check="-mno-evex512"):
                gcc_avx10_flags = ["-mno-evex512"]

            if machine.startswith("riscv"):
                arch_flag = "-march=rv" + ("32" if machine.startswith("riscv32") else "64") + "gc"
                if autoconf_check(self.compiler, flag_check=arch_flag + "v"):
                    gcc_rvv_flags = [arch_flag + "v"]
                if autoconf_check(self.compiler, flag_check=arch_flag + "_zbkc"):
                    gcc_rvzbkc_flags = [arch_flag + "_zbkc"]

        srcdeps_crc_common = ["src/yencode/common.h", "src/yencode/crc_common.h", "src/yencode/crc.h"]
        srcdeps_dec_common = ["src/yencode/common.h", "src/yencode/decoder_common.h", "src/yencode/decoder.h"]
        srcdeps_enc_common = ["src/yencode/common.h", "src/yencode/encoder_common.h", "src/yencode/encoder.h"]

        # build yencode/crcutil
        output_dir = os.path.dirname(self.build_lib)
        compiled_objects = []
        for source_files in [
            {
                "sources": [
                    "src/yencode/platform.cc",
                    "src/yencode/encoder.cc",
                    "src/yencode/decoder.cc",
                    "src/yencode/crc.cc",
                ],
                "include_dirs": ["src/crcutil-1.0/code", "src/crcutil-1.0/examples"],
            },
            {
                "sources": ["src/yencode/encoder_sse2.cc"],
                "depends": srcdeps_enc_common + ["encoder_sse_base.h"],
                "gcc_x86_flags": ["-msse2"],
            },
            {
                "sources": ["src/yencode/decoder_sse2.cc"],
                "depends": srcdeps_dec_common + ["decoder_sse_base.h"],
                "gcc_x86_flags": ["-msse2"],
            },
            {
                "sources": ["src/yencode/encoder_ssse3.cc"],
                "depends": srcdeps_enc_common + ["encoder_sse_base.h"],
                "gcc_x86_flags": ["-mssse3"],
            },
            {
                "sources": ["src/yencode/decoder_ssse3.cc"],
                "depends": srcdeps_dec_common + ["decoder_sse_base.h"],
                "gcc_x86_flags": ["-mssse3"],
            },
            {
                "sources": ["src/yencode/crc_folding.cc"],
                "depends": srcdeps_crc_common,
                "gcc_x86_flags": ["-mssse3", "-msse4.1", "-mpclmul"],
            },
            {
                "sources": ["src/yencode/crc_folding_256.cc"],
                "depends": srcdeps_crc_common,
                "gcc_x86_flags": gcc_vpclmulqdq_flags,
                "msvc_x86_flags": ["/arch:AVX2"],
            },
            {
                "sources": ["src/yencode/encoder_avx.cc"],
                "depends": srcdeps_enc_common + ["encoder_sse_base.h"],
                "gcc_x86_flags": ["-mavx", "-mpopcnt"],
                "msvc_x86_flags": ["/arch:AVX"],
            },
            {
                "sources": ["src/yencode/decoder_avx.cc"],
                "depends": srcdeps_dec_common + ["decoder_sse_base.h"],
                "gcc_x86_flags": ["-mavx", "-mpopcnt"],
                "msvc_x86_flags": ["/arch:AVX"],
            },
            {
                "sources": ["src/yencode/encoder_avx2.cc"],
                "depends": srcdeps_enc_common + ["encoder_avx_base.h"],
                "gcc_x86_flags": ["-mavx2", "-mpopcnt", "-mbmi", "-mbmi2", "-mlzcnt"],
                "msvc_x86_flags": ["/arch:AVX2"],
            },
            {
                "sources": ["src/yencode/decoder_avx2.cc"],
                "depends": srcdeps_dec_common + ["decoder_avx2_base.h"],
                "gcc_x86_flags": ["-mavx2", "-mpopcnt", "-mbmi", "-mbmi2", "-mlzcnt"],
                "msvc_x86_flags": ["/arch:AVX2"],
            },
            {
                "sources": ["src/yencode/encoder_vbmi2.cc"],
                "depends": srcdeps_enc_common + ["encoder_avx_base.h"],
                "gcc_x86_flags": gcc_vbmi2_flags + gcc_avx10_flags,
                "msvc_x86_flags": ["/arch:AVX512"],
            },
            {
                "sources": ["src/yencode/decoder_vbmi2.cc"],
                "depends": srcdeps_dec_common + ["decoder_avx2_base.h"],
                "gcc_x86_flags": gcc_vbmi2_flags + gcc_avx10_flags,
                "msvc_x86_flags": ["/arch:AVX512"],
            },
            {
                "sources": ["src/yencode/encoder_neon.cc"],
                "depends": srcdeps_enc_common,
                "gcc_arm_flags": gcc_arm_neon_flags,
            },
            {
                "sources": ["src/yencode/decoder_neon64.cc" if IS_AARCH64 else "src/yencode/decoder_neon.cc"],
                "depends": srcdeps_dec_common,
                "gcc_arm_flags": gcc_arm_neon_flags,
            },
            {
                "sources": ["src/yencode/crc_arm.cc"],
                "depends": srcdeps_crc_common,
                "gcc_arm_flags": gcc_arm_crc_flags,
            },
            {
                "sources": ["src/yencode/crc_arm_pmull.cc"],
                "depends": srcdeps_crc_common,
                "gcc_arm_flags": gcc_arm_crc_pmull_flags,
            },
            {
                "sources": ["src/yencode/encoder_rvv.cc", "src/yencode/decoder_rvv.cc"],
                "depends": srcdeps_enc_common + srcdeps_dec_common,
                "gcc_flags": gcc_rvv_flags,
            },
            {
                "sources": ["src/yencode/crc_riscv.cc"],
                "depends": srcdeps_crc_common,
                "gcc_flags": gcc_rvzbkc_flags,
            },
            {
                "sources": [
                    "src/crcutil-1.0/code/crc32c_sse4.cc",
                    "src/crcutil-1.0/code/multiword_64_64_cl_i386_mmx.cc",
                    "src/crcutil-1.0/code/multiword_64_64_gcc_amd64_asm.cc",
                    "src/crcutil-1.0/code/multiword_64_64_gcc_i386_mmx.cc",
                    "src/crcutil-1.0/code/multiword_64_64_intrinsic_i386_mmx.cc",
                    "src/crcutil-1.0/code/multiword_128_64_gcc_amd64_sse2.cc",
                    "src/crcutil-1.0/examples/interface.cc",
                ],
                "gcc_flags": ["-Wno-expansion-to-defined", "-Wno-unused-parameter"],
                "include_dirs": ["src/crcutil-1.0/code", "src/crcutil-1.0/tests"],
                "macros": [("CRCUTIL_USE_MM_CRC32", "0")],
            },
            {
                "sources": [
                    "src/yenc.cc",
                ],
                "gcc_flags": ["-Wno-unused-parameter"],
            },
            {
                "sources": [
                    "src/unlocked_ssl.cc",
                ],
                "gcc_flags": ["-Wno-unused-parameter", "-Wno-missing-field-initializers"],
                "msvc_libraries": ["ws2_32"],
            },
            {
                "sources": [
                    "src/crc32.cc",
                ],
                "gcc_flags": ["-Wno-unused-parameter"],
                "include_dirs": ["src/crcutil-1.0/code", "src/crcutil-1.0/examples"],
            },
            {
                "sources": [
                    "src/sparse.cc",
                ],
                "gcc_flags": ["-Wno-unused-parameter"],
            },
            {
                "sources": [
                    "src/utils.cc",
                ],
                "gcc_flags": ["-Wno-unused-parameter"],
            },
        ]:
            args = {
                "sources": source_files["sources"],
                "output_dir": output_dir,
                "extra_postargs": cflags[:],
                "macros": gcc_macros[:],
            }
            if self.compiler.compiler_type == "msvc":
                if IS_X86 and "msvc_x86_flags" in source_files:
                    args["extra_postargs"] += source_files["msvc_x86_flags"]
                if "msvc_libraries" in source_files:
                    ext.libraries += source_files["msvc_libraries"]
            else:
                if "gcc_flags" in source_files:
                    args["extra_postargs"] += source_files["gcc_flags"]
                if IS_X86 and "gcc_x86_flags" in source_files:
                    args["extra_postargs"] += source_files["gcc_x86_flags"]
                if IS_ARM and "gcc_arm_flags" in source_files:
                    args["extra_postargs"] += source_files["gcc_arm_flags"]

            if "include_dirs" in source_files:
                args["include_dirs"] = source_files["include_dirs"]
            if "macros" in source_files:
                args["macros"].extend(source_files["macros"])

            self.compiler.compile(**args)
            compiled_objects += self.compiler.object_filenames(source_files["sources"], output_dir=output_dir)

        # attach to Extension
        ext.extra_link_args = ldflags + compiled_objects
        ext.depends = ["src/sabctools.h"] + compiled_objects

        # proceed with regular Extension build
        super(SABCToolsBuild, self).build_extension(ext)


def get_version():
    """Parse the version from the C sources"""
    with open("src/sabctools.h", "r") as sabctools_h:
        version = re.findall('#define SABCTOOLS_VERSION +"([0-9xA-Z_.]+)"?', sabctools_h.read())[0]
        return version


setup(
    version=get_version(),
    ext_modules=[
        Extension(
            "sabctools.sabctools",
            ["src/sabctools.cc"],
        )
    ],
    cmdclass={"build_ext": SABCToolsBuild},
)
